#include <stdio.h>
#include <assert.h>
#include <getopt.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>   
#include <string.h>
#include <ctype.h>
#include <time.h>    

#include "sensus.h"
#include "bool.h"
#include "config.h"

/* ---- POSIX compliant source ---- */
#define _POSIX_SOURCE 1   

#ifndef NDEBUG
static bool verbose = true; /* == true iff debug mode is selected */
#else
static bool verbose = false;
#endif

/* ---- Dump-mode flag ---- */
static bool dump = false;

/* ---- Give all time-stamps in local timezone ---- */
static bool local = false;

/* ---- Indicates whether we have to create groups of repetitive dives ----- */ 
static bool groups = false;

/* ---- Read-timeout in seconds ---- */
static const unsigned int timeout = 60;

/* ---- 1ft == 0.3048 m ---- */
#define FEET2METER 0.3048

/* ---- 1m == 3.2808399 ft ---- */
#define METER2FEET 3.2808399

/* ---- A constant added to each pressure sample (according to Sensus devel-guide) ---- */
#define SAMPLEADJUST ((float) 13)

/* ---- Number of samples in one line in the <SAMPLE></SAMPLE> section ---- */
#define COLUMNS 6

/* ---- Maxiamal length of a FQN hostname ---- */
#define MAXDOMAINNAMELEN 256 

/* ---- Maximal length of a string holding local time ---- */
#define MAXTIMELEN 64

/* ---- Default minimal number of samples for valid dives (Print all dives)  ---- */
#define DEFMINSIZE 0

/* ---- Default number of seconds for complete desaturation (36h) ---- */
#define DEFDESATTIME (36 * 3600) 

/* ---- Altitude of all dives is sealevel by default ---- */
#define DEFALTITUDE 0 

/* ---- A new datatype for holding the index and time of a dive ---- */
typedef struct _diveinfo_struct_ {
                                  int startIndex;
                                  time_t diveTime;
                                 } DiveInfoEntry;

static int CompareDiveInfoEntry(const void* arg1, const void* arg2)
{
 /* ---- Compare the sensus-times (We will sort on the timestamp) ---- */
 if (((DiveInfoEntry *) arg1) -> diveTime == ((DiveInfoEntry *) arg2) -> diveTime) { 

  /* ---- 'arg1' is equal ---- */
  return 0;

 } else if (((DiveInfoEntry *) arg1) -> diveTime < 
            ((DiveInfoEntry *) arg2) -> diveTime) { 

  /* ----  'arg1' is smaller ---- */
  return -1;

 } else { 

  /* ---- 'arg1' is bigger ---- */
  return 1;

 }

}

static bool getFQDomain(char *domain, int len)
{
 struct hostent* he = NULL;
 
 /* ---- Get the hostname ---- */
 if (gethostname(domain, len - 1) == -1) {
 
  /* ---- Can't get hostname (give an error) ---- */
  domain[0] = '\0';
  return false;
 
 };
 
 /* ---- Use the resolver ---- */
 he = gethostbyname(domain);
 
 /* --- And check gethostbyname-call ---- */
 if (he == NULL) {
 
  /* ---- Can't get FQDN (show an error) ---- */
  domain[0] = '\0';
  return false;
 
 }                                                                    

 /* ---- Now copy result (always terminated with '\0' )---- */
 strncpy(domain, he -> h_name, (size_t) len - 1);
 domain[len - 1] = '\0';

 /* ---- Everthing is ok ---- */
 return true;
 
}

static void PrintHelp(char* prgname)
{
#ifdef HAVE_GETOPTLONG
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: sensusconvert.c,v 1.28 2003/12/26 03:50:37 streit Exp $\n");
 fprintf(stderr, "-h --help                 Print a help text\n");
 fprintf(stderr, "-a --alt                  Set altitude of all dives\n");
 fprintf(stderr, "-s --size                 Minimal number of samples in a valid dive\n");
 fprintf(stderr, "-g --groups               Calculate groups of repetitive dives\n");
 fprintf(stderr, "-t --time                 Set time-interval for complete desaturation");
 fprintf(stderr, "-l --local                Give all time-stamps in local timezone\n");
 fprintf(stderr, "-v --verbose              Explain what is done\n");
#ifndef NDEBUG
 fprintf(stderr, "-d --dump                 Print a hexdump of read data\n");
#endif
#else
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose) 
  fprintf(stderr, "$Id: sensusconvert.c,v 1.28 2003/12/26 03:50:37 streit Exp $\n");
 fprintf(stderr, "-h                        Print a help text\n");
 fprintf(stderr, "-a                        Set altitude of all dives\n");
 fprintf(stderr, "-s                        Minimal number of samples in a valid dive\n");
 fprintf(stderr, "-g                        Calculate groups of repetitive dives\n");
 fprintf(stderr, "-t                        Set time-interval for complete desaturation");
 fprintf(stderr, "-l                        Give all time-stamps in local timezone\n");
 fprintf(stderr, "-v                        Explain what is done\n");
#ifndef NDEBUG
 fprintf(stderr, "-d                        Print a hexdump of read data\n");
#endif
#endif

 /* ---- Terminate program successfully ---- */
 exit(EXIT_SUCCESS);

}

static void AlarmHandler(int dummy)
{
 /* ---- Tell an timeout and quit ---- */
 fprintf(stderr, "WARN: A timeout occured while reading!");
 fprintf(stderr, "Better we quit.\n");
 exit(EXIT_FAILURE);

}

static void HexDumpBuffer(char* buffer, int bufferSize)
{
 const int cols = 16;
 int i;

 /* ---- Walk through whole diveBuffer ---- */
 for (i = 0; i < bufferSize; i++) {

  /* ---- Check if we need an offset ---- */
  if ((i % cols) == 0) fprintf(stderr, "%8.8X: ", i);

  /* ---- Print an entry ---- */
  fprintf(stderr, "%c%c", 
                  digits[(((unsigned long) buffer[i] / HEXBASE) % HEXBASE)],
                  digits[((unsigned long)  buffer[i] % HEXBASE)]);

  /* ---- Check the type of seperator ---- */
  if ((i % cols) < (cols - 1))
   fprintf(stderr, " "); 
  else 
   fprintf(stderr, "\n");

 }

 /* ---- Add a new line ---- */
 fprintf(stderr, "\n\n");

}

static void ReadHandshakeBuffer(char handshakeBuffer[HANDSHAKESIZE])
{
#ifndef NDEBUG
 /* ---- Report the installation of the alarm handler ---- */
 fprintf(stderr, "DEBUG: Installing an alarm handler!\n");
#endif

 /* ---- Set the alarm-handler ---- */
 if (signal(SIGALRM, AlarmHandler) == SIG_ERR) {

  /* ---- Give a warning and quit ---- */
  fprintf(stderr, "WARN: Can't install alarm handler!\n");
  fprintf(stderr, "WARN: Have to quit!\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Set the alarm ---- */
 (void) alarm(timeout);

#ifndef NDEBUG
 /* ---- Report "Start reading the profile" ---- */
 fprintf(stderr, "DEBUG: Start reading the handshakeBuffer\n");
#endif

 /* ---- Read the data from stdin ---- */
 if (read(STDIN_FILENO, handshakeBuffer, HANDSHAKESIZE) != HANDSHAKESIZE) {

  /* ---- We can't read the right number of bytes ---- */
  fprintf(stderr, "WARN: Can't read the handshake-data completely! Have to quit\n");
  exit(EXIT_FAILURE);

 }

#ifndef NDEBUG
 /* ---- Tell that we remove the alarm handler ---- */
 fprintf(stderr, "DEBUG: Remove the alarm handler!\n");
#endif

 /* ---- Remove previously set alarm ---- */
 alarm(0);

 /* ---- In check for dump-mode ---- */
 if (dump) {
 
  /* ---- Write a header for the dump ---- */
  fprintf(stderr, "INFO: Dump of handshake buffer\n");

  /* ---- Dump the handshake-buffer ---- */
  HexDumpBuffer(handshakeBuffer, HANDSHAKESIZE);

 }

}


static void CalcSensusTimeWhileDump(char handshakeBuffer[HANDSHAKESIZE], 
				    time_t* sensusTime)
{
 /* ---- Set 'sensusTime' to epoch ---- */
 *sensusTime = (time_t) 0;

 /* ---- Get the sensus-time ---- */
 *sensusTime = (time_t) ((unsigned long)             (0xff & handshakeBuffer[ 8])) +
                        ((unsigned long) (1 <<  8) * (0xff & handshakeBuffer[ 9])) +
                        ((unsigned long) (1 << 16) * (0xff & handshakeBuffer[10])) +
                        ((unsigned long) (1 << 24) * (0xff & handshakeBuffer[11]));

}

static void ReadDiveBuffer(char diveBuffer[PROFILESIZE])
{
#ifndef NDEBUG
 /* ---- Report the installation of the alarm handler ---- */
 fprintf(stderr, "DEBUG: Installing the alarm handler!\n");
#endif

 /* ---- Set the alarm-handler ---- */
 if (signal(SIGALRM, AlarmHandler) == SIG_ERR) {

  /* ---- Give a warning and quit ---- */
  fprintf(stderr, "WARN: Can't install alarm handler!");
  fprintf(stderr, "WARN: Have to quit!\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Set the alarm ---- */
 (void) alarm(timeout);

#ifndef NDEBUG
 /* ---- Report "Start reading the profile" ---- */
 fprintf(stderr, "DEBUG: Start reading the diveBuffer\n");
#endif

 /* ---- Read the data from stdin ---- */
 if (read(STDIN_FILENO, diveBuffer, PROFILESIZE) != PROFILESIZE) {

  /* ---- We can't read the right number of bytes ---- */
  fprintf(stderr, "WARN: Cannot read the profile completely! Have to quit\n");
  exit(EXIT_FAILURE);

 }

#ifndef NDEBUG
 /* ---- Tell that we remove the alarm handler ---- */
 fprintf(stderr, "DEBUG: Remove the alarm handler!\n");
#endif

 /* ---- Remove previously set alarm ---- */
 alarm(0);

 /* ---- In check for dump-mode ---- */
 if (dump) {
 
  /* ---- Write a header for the dump ---- */
  fprintf(stderr, "INFO: Dump of dive-data buffer\n");

  /* ---- Dump the handshake-buffer ---- */
  HexDumpBuffer(diveBuffer, PROFILESIZE);

 }

}

static void ReadDumpTime(time_t* dumpTime)
{
 char ch; /* Buffer for one char to be read */
 int charsRead; /* Number of bytes read by systemcall 'read' */
 bool foundEndMarker = false; /* Becomes true is we find TIMEENDMARKER */

 /* ---- Read all digits of stored sysTime ---- */
 *dumpTime = (time_t) 0;

 /* ---- Read the first digit of dump time (if possible) ---- */
 if (read(STDIN_FILENO, &ch, sizeof(ch)) < sizeof(ch)) return;

 while (!foundEndMarker) {

  /* ---- Add the new digit (This code is clumsy but hopefully portable) ---- */
  switch (ch) {
   case '0': *dumpTime += 0; 
             break;
   case '1': *dumpTime += 1; 
             break;
   case '2': *dumpTime += 2; 
             break;
   case '3': *dumpTime += 3; 
             break;
   case '4': *dumpTime += 4; 
             break;
   case '5': *dumpTime += 5; 
             break;
   case '6': *dumpTime += 6; 
             break;
   case '7': *dumpTime += 7; 
             break;
   case '8': *dumpTime += 8; 
             break;
   case '9': *dumpTime += 9; 
             break;
   case 'A': /* Implement case insensitivity */
   case 'a': *dumpTime += 10; 
             break;
   case 'B': /* Implement case insensitivity */
   case 'b': *dumpTime += 11; 
             break;
   case 'C': /* Implement case insensitivity */
   case 'c': *dumpTime += 12; 
             break;
   case 'D': /* Implement case insensitivity */
   case 'd': *dumpTime += 13; 
             break;
   case 'E': /* Implement case insensitivity */
   case 'e': *dumpTime += 14; 
             break;
   case 'F': /* Implement case insensitivity */
   case 'f': *dumpTime += 15; 
             break;
   default:
           /* ---- Give an error and quit ---- */
           fprintf(stderr, "WARN: File corrupted! Wrong char in dump-time section\n");
	   exit(EXIT_FAILURE);

  }

  /* ---- Get next possible digit ---- */
  charsRead = read(STDIN_FILENO, &ch, sizeof(char));

  /* ---- Check whether we have reached the end of dump-time or input ---- */
  if ((ch != TIMEENDMARKER) && (charsRead >= sizeof(char))) {

   /* ---- Shift one half-byte left ---- */
   *dumpTime = (*dumpTime) * 16;
 
  } else {

   /* ---- Indicate that we have either completely read the dump-time or the input ---- */
   foundEndMarker = true;   

  }

 }

}

static bool CheckHandshakeSection(char handshakeBuffer[HANDSHAKESIZE])
{
 /* ---- Check for the string OK ---- */
 if ((handshakeBuffer[0] == 'O') && (handshakeBuffer[1] == 'K')) 
  return true;
 else 
  return false;

}

static void WriteAddInfo(char handshakeBuffer[HANDSHAKESIZE])
{
 /* ---- Write ADDINFO opening tag ---- */
 printf("<ADDINFO>\n");

 /* ---- Write the serial number ---- */
 printf(" <SERIALID>%d</SERIALID>\n", 
        ((0xff & handshakeBuffer[7]) << 8) + 
        ( 0xff & handshakeBuffer[6]));

 /* ---- Write the corresponding ADDINFO closing tag ---- */
 printf("</ADDINFO>\n");

}

static bool CheckProfileSection(char diveBuffer[PROFILESIZE])
{
 /* ---- Check for the string DATA ---- */
 if ((diveBuffer[0] == 'D') && (diveBuffer[1] == 'A') &&
     (diveBuffer[2] == 'T') && (diveBuffer[3] == 'A'))
  return true;
 else 
  return false;

}

static int SearchFirstHeader(char diveBuffer[PROFILESIZE])
{
 int searchFrom = PREHEADERSIZE;

 /* ---- Sequentially search for a header-marker (0xff) ---- */
 while ((searchFrom < (DIVEDATASIZE + PREHEADERSIZE)) && 
        (diveBuffer[searchFrom] != MAGICHEADER)) searchFrom++;

 /* ---- Check if we have found a header-marker ---- */
 if (searchFrom >= (DIVEDATASIZE + PREHEADERSIZE)) 
  return -1;
 else
  return searchFrom;

}

static int SearchNextHeader(char diveBuffer[PROFILESIZE], int searchFrom)
{
 /* ---- Sequentially search for a header-marker (0xff) ---- */
 while (diveBuffer[searchFrom] != MAGICHEADER) 
  searchFrom = WRAP(searchFrom + 1);

 /* ---- Return the position found ---- */
 return searchFrom;

}

static struct tm* MakeBrokenTime(const time_t* timeStamp, bool local)
{
 /* ---- Should we use the local time-zone? ---- */
 if (local)
  return localtime(timeStamp);
 else
  return gmtime(timeStamp);

}

static time_t GetTimeOfDive(char diveBuffer[PROFILESIZE], int diveStart)
{
 /* ---- Get the dive time-stamp (measured by Sensus' time-normal) ---- */
 return (time_t) 
        (unsigned long)              (0xff & diveBuffer[WRAPOFF(diveStart, 2)])  +
        (unsigned long) ((1 <<  8) * (0xff & diveBuffer[WRAPOFF(diveStart, 3)])) +
        (unsigned long) ((1 << 16) * (0xff & diveBuffer[WRAPOFF(diveStart, 4)])) +
        (unsigned long) ((1 << 24) * (0xff & diveBuffer[WRAPOFF(diveStart, 5)]));

}

static bool CalcDiveInterval(char diveBuffer[PROFILESIZE], 
                      int diveStart, 
                      int* diveStop, 
                      int* sampleStart,
                      int* sampleStop,
                      int* shallowStart,
                      int minSize)
{
 int shallowCnt = 0;
 int i = 0;

 /* ---- Search end of dive (Don't trust the promise of 17 shallow samples) ---- */
 *sampleStart = diveStart + DIVEHEADERSIZE;
 *sampleStop = *sampleStart - 1;
 *shallowStart = *sampleStop;
 shallowCnt = 0;
 i = 0; 
 while((diveBuffer[WRAPOFF(*sampleStop, 1)] != MAGICHEADER)) {

  /* ---- Look at the next sample ---- */ 
  (*sampleStop)++;

  /* ---- Check for a depth sample (not a temperature) ---- */
  if ((i % 7) != 1) {

   /* ---- Check for shallow depth ---- */
   if ((unsigned char) diveBuffer[WRAP(*sampleStop)] < SHALLOWDEPTH) {

    /* ---- Increase the number of found shallow samples ---- */
    shallowCnt++;

    /* ---- Check if we have found enough shallow samples ---- */
    if (shallowCnt >= SHALLOWCNT) break;
 
   } else {

    /* ---- Reset the shallow samples counter ---- */
    shallowCnt = 0;

    /* ---- Set the possible start of shallow samples to actual sample index ---- */
    *shallowStart = *sampleStop;

   }

  }

  /* ---- Increase sample-counter ---- */
  i++;

 }

 /* ---- If no proper region of shallow samples were found give a warning ---- */
 if (shallowCnt < SHALLOWCNT) {

  /* ---- Give a warning ---- */
  fprintf(stderr, "WARN: Profile data is corrupted (Can't find end of dive correctly)\n");

 } else {

  /* ---- Search for the next MAGICHEADER ---- */
  while(diveBuffer[WRAPOFF(*sampleStop, 1)] != MAGICHEADER) (*sampleStop)++;

 }

 /* ---- Return the interval informations of dive ---- */
 *diveStop = WRAP(*sampleStop + 1);

 /* ----- Check if we have to ignore this dive ---- */
 if ((*shallowStart - *sampleStart + 1) < minSize) {

  /* ---- Check for verbose mode ---- */
  if (verbose) {

   /* ---- Write a message ---- */
   fprintf(stderr, 
           "INFO: Ignore dive from 0x%x to 0x%x (too short)\n", 
           diveStart, 
           *shallowStart);

  }

  /* ---- Indicate that we have to ignore this dive ---- */
  return true;

 } else {

  /* ---- Indicate that we have to process this dive ---- */
  return false;

 }

}

static bool IsValidDive(char diveBuffer[PROFILESIZE], int diveStart, int minSize)
{
 int diveStop, sampleStart, sampleStop, shallowStart;

 /* ---- Return whether the dive is long enough ---- */
 return !CalcDiveInterval(diveBuffer, 
                          diveStart, 
                          &diveStop, 
                          &sampleStart, 
                          &sampleStop, 
                          &shallowStart, 
                          minSize);

}

static int WriteDive(char diveBuffer[PROFILESIZE], 
		     time_t dumpTime, 
		     time_t sensusTime, 
		     int diveStart,
		     int minSize,
		     int altitude,
		     time_t *timeOfOldDive,
		     bool *ignored)
{
 int sampleStart; /* Start of sample section */
 int sampleStop; /* End of sample section */
 int shallowStart; /* Start of shallow samples region */
 int samplingInterval = -1 ;
 int i = -1;
 int diveStop; /* Position where the actual dive stops */
 time_t diveTimeOff = (time_t) 0; /* Time of dive in seconds since 'sensusTime' */
 time_t diveTime = (time_t) 0; /* Time when the dive was done */
 struct tm* brokenTime = NULL; /* Pointer to broken time strcut for 'diveTime' */
 char timestr[MAXTIMELEN + 1] = ""; /* String holding a time-string */
 size_t time_res; /* Result of strftime */

 /* ---- Get the dive interval and return if this is an invalid dive ----- */
 if ((*ignored = CalcDiveInterval(diveBuffer, 
                                  diveStart,  
                                  &diveStop, 
                                  &sampleStart, 
                                  &sampleStop, 
                                  &shallowStart, 
                                  minSize))) return diveStop;

 /* ---- Check for verbose-mode ---- */
 if (verbose) {

  /* ---- Write some memory-offsets ---- */
  fprintf(stderr, "INFO: Write dive from 0x%x to 0x%x\n", diveStart, shallowStart);
  
 }

 /* ---- Write DIVE opening tag ---- */
 printf(" <DIVE>\n");

 /* ---- Get the sampling interval ---- */
 samplingInterval = (int) diveBuffer[WRAPOFF(diveStart, 1)];

 /* ---- Get the dive time-stamp ---- */
 diveTimeOff = GetTimeOfDive(diveBuffer, diveStart);
 diveTime = (time_t) (difftime(dumpTime, difftime(sensusTime, diveTimeOff)) + 0.5);

 /* ---- Convert 'diveTime' to broken-time ---- */
 brokenTime = MakeBrokenTime(&diveTime, local);

 /* ---- Write a comment about the used time-zone ---- */
 time_res = strftime(timestr, MAXTIMELEN, "%Z", brokenTime);
 if (!local)
  printf("  <!-- Date and time are given in Coordinated Universal Time (UTC) -->\n");
 else
  printf("  <!-- Date and time are given for time-zone %s -->\n", timestr);

 /* ---- Write the date section ---- */
 printf("  <DATE>");
 time_res = strftime(timestr, MAXTIMELEN, "%Y", brokenTime);
 printf("<YEAR>%s</YEAR>", timestr);
 time_res = strftime(timestr,MAXTIMELEN, "%m", brokenTime);
 printf("<MONTH>%s</MONTH>", timestr);
 time_res = strftime(timestr, MAXTIMELEN, "%d", brokenTime);
 printf("<DAY>%s</DAY>", timestr);
 printf("</DATE>\n");

 /* ---- Write the time section ---- */
 printf("  <TIME>");
 time_res = strftime(timestr, MAXTIMELEN, "%H", brokenTime);
 if (timestr[0] == '0') {
  printf("<HOUR>%s</HOUR>", &timestr[1]);
 } else {
  printf("<HOUR>%s</HOUR>", timestr);
 } 
 time_res = strftime(timestr, MAXTIMELEN, "%M", brokenTime);
 printf("<MINUTE>%s</MINUTE>", timestr);
 printf("</TIME>\n");

 /* ---- Insert the internal sensus-time ---- */
 printf("  <!-- This date/time is equivalent to the internal sensus time: %lu -->\n", 
        (unsigned long) diveTimeOff);

 /* ---- Check whether this is the beginning of a new rep-group ---- */
 if (*timeOfOldDive == (time_t) 0) { 

  /* ---- The surface-interval ---- */
  printf("  <SURFACEINTERVAL>infinity</SURFACEINTERVAL>\n");

 } else {

  /* ---- The surface-interval ---- */
  printf("  <SURFACEINTERVAL>%lu</SURFACEINTERVAL>\n",
	 (unsigned long) (difftime(diveTimeOff, *timeOfOldDive) + 0.5));

 }

 /* ---- Set time of this dive to 'timeOfOldDive' for next call ---- */
 *timeOfOldDive = diveTimeOff;

 /* ---- Give the desity of sea-water ---- */
 printf("  <DENSITY>1030.000</DENSITY>\n");
 
 /* ---- Altitude is sea-level ---- */
 printf("  <ALTITUDE>%d.0</ALTITUDE>\n", altitude);

 /* ---- Insert a mixes section. Use 'Air' only ---- */
 printf("  <GASES>\n");
 printf("   <MIX>\n");
 printf("    <MIXNAME>RealAir</MIXNAME>\n");
 printf("    <O2>0.21</O2>\n");
 printf("    <N2>0.79</N2>\n");
 printf("    <HE>0.00</HE>\n");
 printf("   </MIX>\n");
 printf("  </GASES>\n");

 /* ---- Tag for the delta - mode ---- */
 printf("  <DELTAMODE/>\n");

 /* ---- Write the samples header ---- */
 printf("  <SAMPLES>\n");

 /* ---- Print sample interval ---- */
 printf("   <DELTA>%d.0</DELTA>\n", samplingInterval);

 /* ---- Switch to air ---- */
 printf("   <SWITCH>RealAir</SWITCH>\n");
 printf("   ");

 /* ---- Iterate all samples (including the last one. Hence "<= shallowStart") ---- */
 for (i = sampleStart; i <= shallowStart; i++) {

  /* ---- Check for pressure/depth-sample ---- */
  if ((i - sampleStart) % 7 != 1) {

   /* ---- Write the depth-sample ---- */
   printf("<D>%6.2f</D>", 
          (((float) (unsigned char) diveBuffer[WRAP(i)]) - SAMPLEADJUST) * FEET2METER);

  } else {

   /* ---- Write the temperature-sample ---- */
   printf("\n   <TEMPERATURE>%6.2f</TEMPERATURE>\n   ", 
          (5.0 / 9.0 * ((float) (unsigned char) diveBuffer[WRAP(i)] - 32.0)));

  }

 }

 /* ---- We go up to the surface ---- */
 printf("\n   <D>  0.00</D>\n   <ALARM>SURFACE</ALARM>\n");

 /* ---- Write the samples closing tag ---- */
 printf("  </SAMPLES>\n");

 /* ---- Write DIVE closing tag ---- */
 printf(" </DIVE>\n");

 /* ---- Return position of next dive ---- */
 return WRAP(sampleStop + 1);

}

static bool WriteAllDives(char diveBuffer[PROFILESIZE], 
			  time_t dumpTime, 
			  time_t sensusTime,
			  int minSize,
			  int desatTime, 
			  int altitude) 
{
 int firstDiveIndex = -1;
 int nextDive = -1;
 int numDives = -1;
 int numValidDives = -1;
 int i;
 DiveInfoEntry* allDives;
 DiveInfoEntry* allValidDives;
 bool ignored = true;
 time_t timeOfOldDive;

 /* ---- Look for the index of the first dive-data in profile dump ---- */
 firstDiveIndex = SearchFirstHeader(diveBuffer);

 /* ---- Check if we have found a header ---- */
 if (firstDiveIndex == -1) {

  /* ---- Give an error and quit ---- */
  fprintf(stderr, "WARN: Profile-data corrupted (No dive-data header found)! ");
  fprintf(stderr, "Better we quit.\n");
  exit(EXIT_FAILURE);

 }

 /* ---- The index of the next dive is set to the first dive ---- */
 nextDive = firstDiveIndex;

 /* ---- Init the dive counter ---- */
 numDives = 0;

 /* ---- Count number of dives in actual dump ---- */
 do {

  /* ---- Search the next dive ---- */
  nextDive = SearchNextHeader(diveBuffer, WRAP(nextDive + DIVEHEADERSIZE));

  /* --- Increment the dive-counter ---- */
  numDives++;

 } while (nextDive != firstDiveIndex); 

 /* ---- Check for verbose mode ---- */
 if (verbose) {

  /* ---- Give the number of dives found ---- */
  fprintf(stderr, "INFO: Number of dives found: %d\n", numDives);

 }

 /* ---- Allocate enough memory for the index/time of all dives ---- */
 allDives = malloc(sizeof(DiveInfoEntry) * (numDives + 1)); 

 /* ---- Check the allocation ---- */
 if (allDives == NULL) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "WARN: No memory for complete dive information. Have to quit!\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Allocate enough memory for the index/time of all valid dives ---- */
 allValidDives = malloc(sizeof(DiveInfoEntry) * (numDives + 1)); 

 /* ---- Check the allocation ---- */
 if (allValidDives == NULL) {

  /* ---- Give an error ---- */
  fprintf(stderr, "WARN: No memory for processing dive information. Have to quit!\n");

  /* ---- Free some alread allocated memory ---- */
  free(allDives);
  allDives = (DiveInfoEntry *) NULL;

  /* ---- Exit with an error ---- */
  exit(EXIT_FAILURE);

 }


 /* ---- The index of the next dive to print is the first dive ---- */
 nextDive = firstDiveIndex;

 /* ---- Initialize the universal index ---- */
 i = 0; 

 do {

  /* ---- Store the dive information ---- */
  allDives[i].startIndex = nextDive;
  allDives[i].diveTime = GetTimeOfDive(diveBuffer, nextDive);

  /* ---- Increment the index ---- */
  i++;

  /* ---- Search the next dive ---- */
  nextDive = SearchNextHeader(diveBuffer, WRAP(nextDive + DIVEHEADERSIZE));

 } while (nextDive != firstDiveIndex);

 /* ---- Copy the information for all valid (having enough samples) dives ---- */
 numValidDives = 0;
 for (i = 0; i < numDives; i++) {

  /* ---- Check if this is a valid dive ---- */
  if (IsValidDive(diveBuffer, allDives[i].startIndex, minSize)) {

   /* ---- Copy the dive information ---- */
   allValidDives[numValidDives] = allDives[i];

   /* ---- Increment the index ---- */
   numValidDives++;

  }

 }

 /* ---- Sort the array ---- */
 qsort(allValidDives, numValidDives, sizeof(DiveInfoEntry), CompareDiveInfoEntry);

 /* ---- Check for verbose mode ---- */
 if (verbose) {

  /* ---- Print the number of valid dives ---- */
  fprintf(stderr, "INFO: Number of valid dives to print: %d\n", numValidDives);

 }

#ifndef NDEBUG
 /* ---- Write information about all dives ---- */
 for (i = 0; i < numValidDives; i++) {

  /* ---- Write the i-th entry in 'allValidDives' ----- */
  fprintf(stderr, 
          "DEBUG: Index %d, Offset 0x%x, Time %ld\n", 
          i, 
          allValidDives[i].startIndex, 
          (long) allValidDives[i].diveTime);

 }
#endif

 /* ---- Write REPGROUP opening tag for the first dive in dump ---- */
 printf("<REPGROUP>\n");

 /* ---- Set the time of previous dive ---- */
 timeOfOldDive = (time_t) 0;

 /* ---- Write all dives stored in 'allDives' ----- */
 for(i = 0; i < numValidDives; i++) {

  /* ---- Dump the dive ----- */
  nextDive = WriteDive(diveBuffer, 
                       dumpTime, 
                       sensusTime, 
                       allValidDives[i].startIndex, 
                       minSize,
		       altitude,
		       &timeOfOldDive,
                       &ignored);

  /* ---- Check whether we have the end of a group of consecutive dives ---- */
  if ((groups) && 
      ((i + 1) < numValidDives) && 
      (difftime(allValidDives[i + 1].diveTime, 
                 allValidDives[i].diveTime) > (double) desatTime)) { 

   /* ---- Write REPGROUP closing/opening tag if we are completely desaturated ---- */
   printf("</REPGROUP>\n<REPGROUP>\n");

   /* ---- Reset time of previous dive to 0 ---- */
   timeOfOldDive = (time_t) 0;

  }

 }

 /* ---- Write REPGROUP closing tag, because we have completely written the dump ---- */
 printf("</REPGROUP>\n");

 /* ---- Free all allocated memory ---- */
 free(allDives);  allDives = NULL;
 free(allValidDives); allValidDives = NULL;

 /* ---- Tell we we havn't dumped any dive ---- */
 return (numValidDives == 0);

}

int main(int argc, char* argv[])
{
 char diveBuffer[PROFILESIZE];
 char handshakeBuffer[HANDSHAKESIZE]; 
#ifdef HAVE_GETOPTLONG
 int optindex = (int) - 1;
 struct option longargdesc[] = {
                                {"verbose", no_argument,       NULL, (int)'v'},
                                {"help",    no_argument,       NULL, (int)'h'},
                                {"dump",    no_argument,       NULL, (int)'d'},
				{"alt",     required_argument, NULL, (int)'a'},
				{"size",    required_argument, NULL, (int)'s'},
                                {"groups",  no_argument,       NULL, (int)'g'},
                                {"local",   no_argument,       NULL, (int)'l'},
                                {0,0,NULL,0}
                               };
#endif
 int key = (int) -1;
 int desatTime = DEFDESATTIME; /* Number of seconds for complete desaturation */
 int altitude = DEFALTITUDE; /* Altitude of all dives */
 bool empty = false; /* == false when now dive can be written */
 int minSize = DEFMINSIZE; /* Minimal number of samples for valid dives  */ 
 time_t sysTime; /* Actual system-time */
 time_t dumpTime; /* System-time when the dump was done */
 time_t sensusTime; /* Internal sensus-time when the dump was done */
 char timestr[MAXTIMELEN + 1] = ""; /* Timestring */
 size_t time_res; /* Result of strftime */
 char domainname[MAXDOMAINNAMELEN + 1] = ""; /* Buffer for getdomainname */
 char *log_name; /* Pointer to buffer with login name */                        
 char *argPtr; /* Pointer to name of this program */

 /* ---- Get name of program ---- */
 argPtr = strrchr(argv[0], '/');
 if (argPtr != NULL) {

  /* ---- Pointer to name of program is set to one char behind the slash ---- */ 
  argPtr++;

 } else {

  /* ---- No slash found, hence the name is 'argv[0]' ---- */
  argPtr = argv[0];

 }

 /* ---- Parse all options ---- */
#ifdef HAVE_GETOPTLONG
 while ((key = getopt_long(argc, 
                           argv, 
                           "vhda:s:gl", 
                           longargdesc, 
                           &optindex)) != EOF) {
#else
 while ((key = getopt(argc, argv, "vhda:s:gl")) != EOF) {
#endif
 
  /* ---- Check for specific option ---- */
  switch (key) {
 
   case 'v':
            /* ---- Set verbose mode ---- */
            verbose = true;

            break; 
   case 'a':
            /* ---- Set the altitude for all dives ---- */
            altitude = atoi(optarg);

            /* ---- Check for verbose mode ---- */
            if (verbose) {

             /* ---- Give some info's about altitude mode ---- */
             fprintf(stderr, "INFO: Set altitude to %d\n", altitude);

	    }

            break; 
   case 'd':
            /* ---- Set hexdump mode ---- */
            dump = true;

            break; 
   case 'g':
            /* ---- Set groups - mode (We calculate repetitive-dive groups) ---- */
            groups = true;

            /* ---- Check for verbose mode ---- */
            if (verbose) {
             
             /* ---- Report that we generate rep-groups ---- */
             fprintf(stderr, "INFO: We calculate repetitive-dive groups\n");

	    }

            break; 
   case 't':
            /* ---- Set the time-interval for complete desaturation ---- */
            desatTime = atoi(optarg);

            /* ---- Check for verbose mode ---- */
            if (verbose) {

             /* ---- Write the desaturation time used ---- */
             fprintf(stderr, "INFO: Set the desaturation time to %d\n", desatTime);

	    }

            break; 

   case 's':
            /* ---- Set the minimal size ---- */
            minSize = atoi(optarg);

            /* ---- Check for verbose mode ---- */
            if (verbose) {

	     /* ---- Write the minimal size of a valid dive ---- */
             fprintf(stderr, "INFO: Set minimal size of a valid dive to %d\n", minSize);

	    }

            break; 
   case 'l':
            /* ---- Convert all time-stamps to local time-zone ---- */
            local = true;

            /* ---- Check for verbose mode ---- */
            if (verbose) {

	     /* ---- Tell something about the time-mode ---- */
             fprintf(stderr, "INFO: All dates will be given in local time\n");

	    }

            break; 
   case 'h':
            /* ---- Print the help message ---- */
            PrintHelp(argv[0]);

            break;
   case '?':

#ifdef HAVE_GETOPTLONG
            printf("Try `%s --help' for more information\n", argv[0]);
#else
            printf("Try `%s -h' for more information\n", argv[0]);
#endif
            exit(EXIT_FAILURE);

   default:
            printf("Error while parsing options!\n");
            exit(EXIT_FAILURE);

  }

 }

 /* ---- Get the system-time when the dump was done ---- */
 ReadDumpTime(&dumpTime);

 /* ---- Read the handshake data from stdin ---- */
 ReadHandshakeBuffer(handshakeBuffer);

 /* ---- Check the header of the handshake-section ---- */
 if (!CheckHandshakeSection(handshakeBuffer)) {

  /* ---- Give an error and quit ---- */
  fprintf(stderr, "WARN: Profile-data corrupted (Incorrect handshake-section)! ");
  fprintf(stderr, "Better we quit.\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Get the sensus-time when the dump was read ---- */
 CalcSensusTimeWhileDump(handshakeBuffer, &sensusTime);

 /* ---- Print the Header ---- */
 printf("<PROFILE UDCF=\"1\">\n");

 /* ---- Write some header information ---- */
 printf("<!-- This file was generated by '%s' -->\n", argPtr);
 
 /* ---- Get login name ---- */
 log_name = getenv("LOGNAME");

 /* ---- Check for valid login name and hostname ---- */
 if (log_name != NULL) {
 
  /* ---- Get the FQN hostname ---- */
  if(getFQDomain(domainname, MAXDOMAINNAMELEN)) {
 
   /* ---- Write login/hostname/domainname ---- */
   printf("<!-- Converted for: %s@%s -->\n", log_name, domainname);
 
  } else {
 
   /* ---- Write login/hostname ---- */
   printf("<!-- Converted for: %s -->\n", log_name);
 
  }
 
 }                                               

 /* ---- Get system-time ---- */
 sysTime = time(NULL);
 time_res = strftime(timestr,
                     MAXTIMELEN,
                     "%a %b %d %H:%M:%S %Z %Y",
                     MakeBrokenTime((time_t *) &sysTime, local)
                    );

 /* ---- Insert conversion time in header ---- */
 printf("<!-- Date: %s -->\n", timestr);

 /* ---- Convert dump-time ---- */
 time_res = strftime(timestr,
                     MAXTIMELEN,
                     "%a %b %d %H:%M:%S %Z %Y",
                     MakeBrokenTime((time_t *) &dumpTime, local));

 /* ---- Insert dump-time in header ---- */
 printf("<!-- Dumped: %s -->\n", timestr);

 /* ---- Insert the internal sensus-time ---- */
 printf("<!-- Sensus Ticks: %lu -->\n", (unsigned long) sensusTime);
 
 /* ---- We will write all units in metric system ---- */
 printf("<UNITS>Metric</UNITS>\n");

 /* ---- Print Sensus - device info ---- */
 printf("<DEVICE>\n");
 printf(" <!-- See: http://www.reefnet.on.ca/ -->\n");
 printf(" <VENDOR>Reefnet</VENDOR>\n");
 printf(" <MODEL>SENSUS</MODEL>\n");
 printf(" <VERSION>%c</VERSION>\n", handshakeBuffer[3]);
 printf("</DEVICE>\n");

 /* ---- Write the additional info ---- */
 WriteAddInfo(handshakeBuffer);

 /* ---- Read the dive-data from stdin ---- */
 ReadDiveBuffer(diveBuffer);

 /* ---- Check the header of the profile-section ---- */
 if (!CheckProfileSection(diveBuffer)) {

  /* ---- Give an error and quit ---- */
  fprintf(stderr, "WARN: Profile-data corrupted (Incorrect profile-section)! ");
  fprintf(stderr, "Better we quit.\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Write all dive information ---- */
 empty = WriteAllDives(diveBuffer, dumpTime, sensusTime, minSize, desatTime, altitude);
 
 /* ---- Check if we wrote no dive at all ---- */
 if (empty) {

  /* ----- Write a dummy-dive and a warning ---- */
  printf("<REPGROUP>\n");
  printf(" <DIVE>\n");
  printf("  <DATE><YEAR>2001</YEAR><MONTH>08</MONTH><DAY>27</DAY></DATE>\n");
  printf("  <TIME><HOUR>22</HOUR><MINUTE>32</MINUTE></TIME>\n");
  printf("  <SURFACEINTERVAL>infinity</SURFACEINTERVAL>\n");
  printf("  <DENSITY>1030.000</DENSITY>\n");
  printf("  <ALTITUDE>700.0</ALTITUDE>\n");
  printf("  <GASES>\n");
  printf("   <MIX>\n");
  printf("    <MIXNAME>RealAir</MIXNAME>\n");
  printf("   <O2>0.21</O2>\n");
  printf("    <N2>0.79</N2>\n");
  printf("    <HE>0.00</HE>\n");
  printf("   </MIX>\n");
  printf("  </GASES>\n");
  printf("  <DELTAMODE/>\n");
  printf("  <SAMPLES>\n");
  printf("   <DELTA>10.0</DELTA>\n");
  printf("   <SWITCH>RealAir</SWITCH>\n");
  printf("   <D>1.00</D><D>1.00</D>\n");
  printf("   <ALARM>Surface</ALARM>\n");
  printf("  </SAMPLES>\n");
  printf(" </DIVE>\n");
  printf("</REPGROUP>\n");

 }

 /* ---- Write the closing tag ---- */
 printf("</PROFILE>\n");

 /* ---- Everything is fine ---- */
 exit(EXIT_SUCCESS);

}
