#include <assert.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>

#include "sensus.h"
#include "bool.h"
#include "config.h"

#ifndef NDEBUG
#include <ctype.h>
#endif

/* ---- POSIX compliant source ---- */
#define _POSIX_SOURCE 1   

/* ---- Defs for name of serial device, lockfile creation and output filename ---- */
#define DEFDEVICE "/dev/ttyS1"
#define DEFOUTFILENAME "profile.sns"
#define LOCKFILEPROTO "/var/lock/LCK..%s"
#define MAXLOCKLEN ((FILENAME_MAX) + 32) /* Longer than the string in LOCKFILEPROTO */

/* ---- Some time constants (in ms) ---- */
#define LONGWAIT 1000
#define PREHANDSHAKEWAIT 100
#define PREDUMPWAIT 1000

/* ---- Special command-code for the Sensus ---- */ 
#define SENSUS_ALOHA    0x00 /* Wake up Sensus                               */
#define SENSUS_ABORT    0x00 /* We need no dump after a successful handshake */
#define SENSUS_SENDDATA 0x40 /* Send profile data                            */

/* ---- Global variable for verbose mode ---- */
#ifndef NDEBUG
bool verbose = true; /* == true iff verbose mode is selected */
#else 
bool verbose = false;
#endif 

/* ---- Maximal length of character representation of a pid ---- */
#define MAXPIDLEN 32

/* ---- Datatype for the phase of comunication with sensus (handshake / transfer) ---- */
typedef enum { PhaseIllegal,       /* Invalid value                                    */
               PhaseNotContacted,  /* Sensus not contacted yet                         */
               PhaseContacted,     /* Sensus is contacted but handshake is not started */
               PhaseHandshake,     /* Sensus is contacted and we did the handshake     */
               PhaseTransfer,      /* Sensus tries to transfer data at the moment      */
               PhaseFinish         /* Senus has completed the data transfer            */
             } Phase;

/* ---- Filehandle of device and lockfile ---- */
typedef struct _prgstate {
                          /* ---- Name and filehandle of serial device ---- */
                          char devicename[FILENAME_MAX + 1];
                          int device; 

                          /* ---- Name and filehandle of lock file ---- */
                          char lockfile[MAXLOCKLEN + 1];
                          int lockfd;
  
                          /* ---- Name and filehandle of default outputfile ---- */ 
                          char outfilename[FILENAME_MAX + 1];                        
                          int outfd;

                          /* ---- Original status of serial device ---- */
                          struct termios origtermbuf;

                          /* ---- Actual status of serial device ---- */
                          struct termios termbuf; /* Handle for termios functions */

                          /* ---- Actual phase of program (handshake, transfer) ---- */
                          Phase phase;

                          /* ---- Memory for the handshake header ---- */
                          char handShakeData[HANDSHAKESIZE + 1];

                          /* ---- Memory for the profile data ---- */
                          char profileData[PROFILESIZE + 1];

                          /* ---- Current system time ---- */
                          time_t currentTime;

                         } PrgState;

/* ---- Contains all relevant data about this application ---- */
/* ---- It has to be global for the exit-handler          ---- */
static PrgState state;

/* ---- Init the state variable ---- */
void InitState(PrgState* state)
{
 int i;

 /* ---- Set useful default values ---- */
 strcpy(state -> devicename, DEFDEVICE); state -> device = -1;
 state -> lockfile[0] = '\0'; state -> lockfd = -1;
 state -> outfilename[0] = '\0'; state -> outfd = -1;
 strcpy(state -> outfilename, "-");
 state -> phase = PhaseIllegal;

 /* ---- Erase all invalid handshake data ---- */
 for (i = 0; i < HANDSHAKESIZE; i++) state -> handShakeData[i] = '\0';
 
 /* ---- Erase all invalid profile data ---- */
 for (i = 0; i < PROFILESIZE; i++) state -> profileData[i] = '\0';

}

#ifndef NDEBUG
void PrintState(PrgState* state)
{
 /* ---- Print values of devicename/device to stderr ---- */
 if (state -> devicename[0] != '\0')  
  fprintf(stderr, "DEBUG: Devicename: %s\n", state -> devicename);
 else
  fprintf(stderr, "DEBUG: Devicename: undefined\n");
 fprintf(stderr, "DEBUG: Device FD: %d\n", state -> device);

 /* ---- Print value of lockfile to stderr ---- */
 if (state -> lockfile[0] != '\0')  
  fprintf(stderr, "DEBUG: Lockfile: %s\n", state -> lockfile);
 else
  fprintf(stderr, "DEBUG: Lockfile: undefined\n");
 fprintf(stderr, "DEBUG: Lockfile FD: %d\n", state -> lockfd);

}
#endif

/* ---- Suspend execution for 'msec' sec * 10^-3 ---- */ 
void MWait(long msec)
{
 struct timespec ts;

 /* ---- Init timespec ---- */
 ts.tv_sec = (time_t) msec / 1000;
 ts.tv_nsec = (msec % 1000 ) * 1000 * 1000; 

 /* ---- Wait 'msec' sec*10^-3 ---- */
 (void) nanosleep(&ts, (struct timespec *) NULL);

}

/* ---- Create a lockfile for serial device ---- */
void CreateLockfile(PrgState* state)
{
 char mypid[MAXPIDLEN + 1];
 char* slashptr;
 struct stat statbuf;
 mode_t oldmask;

 /* ---- Get name of device without any path information ---- */
 if ((slashptr = strrchr(state -> devicename, '/')) != (char*) NULL) {

  /* ---- Check for valid string after last '/' in 'devicename' ---- */
  if (*(slashptr + 1) != '\0') {

   /* ---- Create the lockfilename (does contain '/') ---- */
   snprintf(state -> lockfile, MAXLOCKLEN - 1, LOCKFILEPROTO, slashptr + 1);

  } else {

  /* ---- Give an error message and terminate ---- */
  fprintf(stderr, "WARN: Empty devicename in %s after last '/'\n", state -> devicename);
  exit(EXIT_FAILURE);

  }

 } else {

  /* ---- Create the lockfilename (doesn't contain '/') ---- */
  snprintf(state -> lockfile, MAXLOCKLEN - 1, LOCKFILEPROTO, state -> devicename);

 }

 /* ---- Check if lockfile already exists ---- */
 if (stat(state -> lockfile, &statbuf) == 0) {

  /* ---- Give an error message and terminate ---- */
  fprintf(stderr, "WARN: Device %s already locked\n", state -> devicename);
  exit(EXIT_FAILURE);

 } 

 /* ---- Check whether we should report this ---- */
 if (verbose) {

  /* ---- Print a message ---- */
  fprintf(stderr, "INFO: Try to establish lock file %s\n", state -> lockfile);

 }

 /* ---- Open the lockfile ---- */
 oldmask = umask(0);
 if ((state -> lockfd = open(state -> lockfile,
                             O_RDWR | O_CREAT, 
                             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: Can't open %s as lockfile. Have to quit.\n", state -> lockfile);
  exit(EXIT_FAILURE);

 }
 (void) umask(oldmask);
 
 /* ---- Write the pid of this process into the lockfile ---- */
 mypid[0] = '\0';
 snprintf(mypid, MAXPIDLEN - 1, "%ld", (long) getpid());
 if (write(state -> lockfd, mypid, strlen(mypid)) != (ssize_t) strlen(mypid)) {

  /* ---- Give an error-message ---- */
  fprintf(stderr, "WARN: Can't write lockfile correctly. Have to quit\n");

  /* ---- Exit with an error ---- */
  exit(EXIT_FAILURE);

 }  

}

/* ---- Restore status to old settings ---- */
void SetDevice(PrgState* state)
{
 /* ---- Check for verbose mode ---- */
 if (verbose) 
  fprintf(stderr, "INFO: Try to open device %s\n", state -> devicename);

 /* ---- Open serial device (Read/write and no blocking)---- */
 state -> device = open(state -> devicename, O_RDWR | O_NOCTTY); /* | O_NONBLOCK); */
 if (state -> device < 0) {

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: Can't open %s as input.\n", state -> devicename);
  exit(EXIT_FAILURE);

 }

 /* ---- Get device status ---- */
 if (tcgetattr(state -> device, &(state -> origtermbuf)) < 0) {

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: Can't get status of %s.\n", state -> devicename);
  exit(EXIT_FAILURE);

 }

 /* ---- Check for verbose mode ---- */
 if (verbose) 
  fprintf(stderr, "INFO: Try to set device %s to 19200 8N1\n", state -> devicename);

 /* ---- Set termios-struct to raw mode and 19200 Baud 8N1 ---- */
 state -> termbuf.c_cflag = B19200 | CS8 | CLOCAL | CREAD;
 state -> termbuf.c_iflag = IGNPAR;
 state -> termbuf.c_oflag = 0;
 state -> termbuf.c_lflag = 0;
 state -> termbuf.c_cc[VTIME] = 1; /* Wait no more than 1/10 sec for data */
 state -> termbuf.c_cc[VMIN] = 1;  /* Read at least one kbyte */

 /* ---- Set the device to raw mode (Mode: 19200 8N1) ---- */
 (void) tcflush(state -> device, TCIFLUSH);
 if (tcsetattr(state -> device, TCSANOW, &(state -> termbuf)) < 0) {

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: Can't set modus of %s\n", state -> devicename);
  exit(EXIT_FAILURE);

 }

 /* ---- Now we have contacted the Sensus ---- */
 state -> phase = PhaseContacted;

}

/* ---- Create the outputfile (if needed) ---- */
void HandleOutput(PrgState* state)
{
 mode_t oldmask;

 /* ---- Check if we have to use stdout ---- */
 if ((strcmp(state -> outfilename, "-") != 0) || 
     (state -> outfilename[0] != '\0')) {

  /* ---- Open outputfile ---- */
  oldmask = umask(0);
  if ((state -> outfd = open(state -> outfilename, 
       O_WRONLY | O_CREAT,
       S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {

   /* ---- Give an error message and exit ---- */
   fprintf(stderr, 
           "WARN: Can't open %s for output. Have to quit.\n", 
           state -> outfilename);
   exit(EXIT_FAILURE);

  }
  (void) umask(oldmask); 

  /* ---- Check for verbose mode ---- */
  if (verbose)
   fprintf(stderr, 
           "INFO: Use %s as output file for profile data\n", 
           state -> outfilename);
  
 } else {

  /* ---- We use stdout ---- */
  state -> outfd = STDOUT_FILENO;

  /* ---- Check for verbose mode ---- */
  if (verbose) 
   fprintf(stderr, "INFO: Use 'stdout' for profile data\n");

 }

}

/* ---- Restore status to old settings ---- */
void RestoreDevice(PrgState* state)
{
 /* ---- Check if device were opened ---- */
 if (state -> device > 0) {

  /* ---- Check whether we should report this ---- */
  if (verbose)
   fprintf(stderr, "INFO: Restore device '%s' to old settings\n", state -> devicename);

  /* ---- Set the device to old mode ---- */
  if (tcsetattr(state -> device, TCSADRAIN, &(state -> origtermbuf)) < 0) {

   /* ---- Give an error message ---- */
   fprintf(stderr, "WARN: Can't restore modus of %s\n", state -> devicename);
   exit(EXIT_FAILURE);

  }

 }

}

/* ---- Handle signals ---- */
void CleanUp()
{
 /* ---- Finish the program ---- */
 exit(0);

}

/* ---- Handle cleanup when exit() was called ---- */
void DoExit()
{
 char command;

 /* ---- Check if we have to send a disconnect message ---- */
 if (state.phase == PhaseHandshake) {

  /* ---- Check for verbose-mode ---- */
  if (verbose)
   fprintf(stderr, "INFO: We will send an abort-command to the connected SENSUS\n");

  /* ---- According to the documentation we send a cancel-dump request ---- */
  command = SENSUS_ABORT;
  if (write(state.device, &command , sizeof(command)) != sizeof(command)) { 

   /* ---- Give an error message ---- */
   fprintf(stderr, "FATAL: Serious error: Can't cancel the dump!\n");

  }

 }

 /* ---- Restore old term modus ---- */
 RestoreDevice(&state);

 /* ---- Check for opened device ---- */
 if (state.device > 0) {

  /* ---- Check for verbose mode ---- */
  if (verbose) 
   fprintf(stderr, "INFO: Closing the device (%s)\n", state.devicename);

  /* ---- Close the device ---- */
  (void) close(state.device);
  state.device = -1;

 }

 /* ---- Check if we can close and unlink the file ---- */
 if (state.lockfd > 0) {
 
  /* ---- Check for verbose mode ---- */
  if (verbose)
   fprintf(stderr, "INFO: Close and unlink lock file %s\n", state.lockfile);

  /* ---- Close the lockfile ---- */
  (void) close(state.lockfd);
  state.lockfd = -1;

  /* ---- Delete the lockfile ----- */
  if (unlink(state.lockfile) < 0) {

   /* ---- Give a warning ---- */
   fprintf(stderr, "FATAL: Can't unlink %s\n", state.lockfile);
   fprintf(stderr, "FATAL: Delete the lockfile (%s) manually\n", state.lockfile); 

  } 

 }

}

/* ---- Print Help-Message ---- */
void PrintHelp(const char* prgname)
{
#ifdef HAVE_GETOPTLONG
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose) 
  fprintf(stderr, "$Id: sensusdump.c,v 1.21 2003/12/26 03:48:55 streit Exp $\n");
 fprintf(stderr, "-h --help                   Print a help text\n");
 fprintf(stderr, "-v --verbose                Explain what is done\n");
 fprintf(stderr, "-d --device filename        Specify seriell device\n");
 fprintf(stderr, "-o --output[=filename | =-] Write read data to specified file\n");
 fprintf(stderr, "\n");
#else
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose) 
  fprintf(stderr, "$Id: sensusdump.c,v 1.21 2003/12/26 03:48:55 streit Exp $\n");
 fprintf(stderr, "-h                        Print a help text\n");
 fprintf(stderr, "-v                        Explain what is done\n");
 fprintf(stderr, "-d filename               Specify seriell device\n");
 fprintf(stderr, "-o [=filename | =-]       Write read data to specified file\n");
 fprintf(stderr, "\n");
#endif

 /* ---- Terminate program successfully ---- */
 exit(EXIT_SUCCESS); 

}

void MakeHandshake(PrgState* state)
{
 const struct timeval oneSec = {1, 0}; /* Specifics 1 sec */
 int retryCount = 0;
 char command;
 long nread = 0;
 long readBytes;
 char charBuff = '\0';
 fd_set readfds; /* Needed to wait for sensus */
 struct timeval timeout; 
 int selectRes;

 /* ---- Init the data-structures needed for reading data via 'select' ---- */
 FD_ZERO(&readfds);
 FD_SET(state -> device, &readfds);

 /* ---- Check for verbose mode ---- */
 if (verbose) 
  fprintf(stderr, "INFO: Start handshake\n");

 /* ---- Check whether we have reached the maximal number of retries ---- */
 while ((retryCount < MAXRETRY) && (state -> phase != PhaseHandshake)) {

  /* ---- Increment the retry-counter ---- */
  retryCount++;

  /* ---- Check for verbose mode and display the number of retries ---- */
  if (verbose) 
   fprintf(stderr, "INFO: Retry No %d\n", retryCount);

  /* ---- Wait a little bit before next try. For the first time this is no problem ---- */
  MWait(PREHANDSHAKEWAIT);

  /* ---- Ask the Sensus to start the handshake ---- */
  command = SENSUS_ALOHA;
  if (write(state -> device, &command , sizeof(command)) != sizeof(command)) { 

   /* ---- Give an error message ---- */
   fprintf(stderr, "WARN: Write error while start of handshake phase!\n");

   /* ---- Terminate program ---- */
   exit(EXIT_FAILURE);

  }

  /* ---- Check for verbose mode and tell that we try to do the handshake ---- */
  if (verbose) fprintf(stderr, "INFO: Sent request for handshake\n");

  /* ---- Wait for the answer (Wait at most 'timeout') ---- */
  timeout = oneSec;
  selectRes = select(state -> device + 1, &readfds, NULL, NULL, &timeout);

  /* ---- Check for an timeout ----- */
  if (selectRes == 0) {

   /* ---- Check for verbose mode and report the timeout ---- */
   if (verbose) 
    fprintf(stderr, "INFO: Timeout! Please check your cable!\n");
   
   /* ---- Proceed with the next retry ---- */
   continue;

  }

  /* ---- Check for an error ----- */
  if (selectRes < 0) {

   /* ---- Give an error ---- */
   fprintf(stderr, 
           "WARN: Error while waiting for SENSUS (Reason %s)! -- Will be ignored\n", 
           strerror(errno));
   
   /* ---- Proceed with the next retry ---- */
   continue;

  }

  /* ---- Read the header (use a simple DFA for this job) ---- */
  while (nread < HANDSHAKESIZE) {

   /* ---- Try to read the next byte ---- */ 
   readBytes = read(state -> device, &(charBuff), sizeof(char));

   /* ---- Check if we got the byte ---- */
   if (readBytes > 0) {
   
    /* ---- Check for garbage-check-mode  ---- */
    if (nread < 2) {

     /* ---- Check for the beginning 'O'. Note that this is a simple      ---- */
     /* ---- finite automaton accepting roughly the string OK[0x00-0xff]+ ---- */ 
     if ((nread == 0) && (charBuff == 'O')) { 

      /* ---- Store the leading 'O' ---- */
      state -> handShakeData[0] = charBuff; 
      nread = 1;
   
     } else if ((nread == 1) && (charBuff == 'K')) {

      /* ---- Store the following 'K' ---- */
      state -> handShakeData[1] = charBuff; 
      nread = 2;

     } else {

      /* ---- Up to now we have read garbage. Ignore it ---- */
      nread = 0;

     }  

    } else {

     /* ---- Simply store this character ---- */
     state -> handShakeData[nread] = charBuff;
     nread++;

    }  

   }

   /* ---- Check for verbose mode and show some progress ---- */
   if (verbose) fprintf(stderr, ".");

  }

  /* ---- Now we are in the handshake-phase ---- */
  state -> phase = PhaseHandshake;

 }

 /* ---- Get the number of seconds since epoch ---- */
 state -> currentTime = time(NULL);

 /* ---- Check for success ---- */
 if (state -> currentTime == (time_t) -1) {

  /* ---- Give a warning message ---- */
  fprintf(stderr, "WARN: Can't get the system time. ");
  fprintf(stderr, "WARN: All dates can't be calculated correctly!\n");

 }

 /* ---- In verbose mode print a newline for the following dots ---- */
 if (verbose)
  fprintf(stderr, "\n");

 /* ---- Check for an unsuccessful try to contact our sensus ---- */
 if ((retryCount >= MAXRETRY) && (state -> phase != PhaseHandshake)) {

  /* ---- Give an error message ---- */
  fprintf(stderr, 
          "WARN: Your Sensus does not answer. Maybe battery is flat! Check cable!\n");

  /* ---- Terminate program ---- */
  exit(EXIT_FAILURE);

 }

}

/* ---- Read profile from the sensus ---- */
void ReadProfile(PrgState* state)
{
 char command;
 char charBuff;
 long nread;
 long readBytes;
 long nBytes;

 /* ---- Check for verbose mode ---- */
 if (verbose)
  fprintf(stderr, "INFO: Start transfer of profile\n");

 /* ---- All device-operations don't wait at more 1 sec ---- */
 state -> termbuf.c_cc[VTIME] = 10; 

 /* ---- Set the new device-mode ---- */
 (void) tcflush(state -> device, TCIFLUSH);
 if (tcsetattr(state -> device, TCSANOW, &(state -> termbuf)) < 0) {

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: ReadProfile can't set modus of %s\n", state -> devicename);
  exit(EXIT_FAILURE);

 }

 /* ---- Tell Sensus to start data transfer ---- */
 command = SENSUS_SENDDATA;
 if (write(state -> device, &command, sizeof(command)) != sizeof(command)) { 

  /* ---- Give an error message ---- */
  fprintf(stderr, "WARN: Can't tell Sensus to start data transfer\n");

  /* ---- Terminate program ---- */
  exit(EXIT_FAILURE);

 }

 /* ----- Now we are in transfer-mode ---- */
 state -> phase = PhaseTransfer;

 /* ---- Wait a little bit ---- */
 MWait(PREDUMPWAIT);

 /* ---- Try to read the profile header (use a DFA for this job) ---- */
 nread = 0;
 while (nread < PREHEADERSIZE) {

  /* ---- Read one additional byte ---- */
  readBytes = read(state -> device, &(charBuff), sizeof(charBuff));

  /* ---- Check whether we have read some data ---- */
  if (readBytes > 0) {

   /* ---- Check for the beginning 'D'. Note that this is a simple        ---- */
   /* ---- finite automaton accepting roughly the string DATA[0x00-0xff]+ ---- */ 
   if ((nread == 0) && (charBuff == 'D')) { 

    /* ---- Store the leading 'D' ---- */
    state -> profileData[nread] = charBuff; 
    nread = 1;

   } else if ((nread == 1) && (charBuff == 'A')) {

    /* ---- Store the following 'A' ---- */
    state -> profileData[nread] = charBuff; 
    nread = 2;

   } else if ((nread == 2) && (charBuff == 'T')) {

    /* ---- Store the following 'T' ---- */
    state -> profileData[nread] = charBuff; 
    nread = 3;

   } else if ((nread == 3) && (charBuff == 'A')) {

    /* ---- Store the following 'A' ---- */
    state -> profileData[nread] = charBuff; 
    nread = 4;

   } else {

    /* ---- Up to now we have read garbage. Ignore it ---- */
    nread = 0;

   }  

   /* ---- Show some progress in verbose mode ---- */
   if (verbose)
    fprintf(stderr, ".");

  }

 }

 /* ---- Check for verbose mode ---- */
 if (verbose) {

  /* ---- Print the message ---- */
  fprintf(stderr, "\nINFO: Profile header found\n");

 }

 /* ---- Read the rest of profile ---- */
 while (nread < PROFILESIZE) {

  /* ---- Check how much data can still be read ---- */
  if (nread + BLOCKSIZE < PROFILESIZE) 
   nBytes = BLOCKSIZE;
  else
   nBytes = PROFILESIZE - nread;

  /* ---- Try to read the next data block ---- */
  readBytes = read(state -> device, &(state -> profileData[nread]), nBytes);

  /* ---- Check if we got the data ---- */
  if (readBytes > 0) {

   /* ---- Set read-index to correct position ---- */
   nread += readBytes;   

   /* ---- Show some progress in verbose mode ---- */
   if (verbose)
    fprintf(stderr, ".");

  } 
  
 }

 /* ---- Check for verbose mode and print a newline ---- */
 if (verbose)
  fprintf(stderr, "\n");

 /* ---- Check for verbose mode ---- */
 if (verbose)
  fprintf(stderr, "INFO: Profile read successfully\n");

 /* ---- Now we are finished. All data is read ---- */
 state -> phase = PhaseFinish;

}

void WriteTime(int outfd, time_t outTime)
{
 int actDigit;

 /* ---- Check for correct argument ---- */
 if (outTime < 0) actDigit = HEXBASE;

 /* ---- time_t can be a long, int or double. Check the existence of next digit ---- */
 if ((double) outTime >= (double) HEXBASE) {

  /* ---- Write the beginning of outTime ---- */
  WriteTime(outfd, (outTime / HEXBASE));

 }

 /* ---- Get the rightmost digit ---- */
 actDigit = ((long) ((double) outTime + 0.5)) % HEXBASE; 

 /* ---- Write the digit ---- */
 if (write(outfd, &(digits[actDigit]), sizeof(char)) != sizeof(char)) {

  /* ---- Tell that outputfile isn't complete ---- */
  fprintf(stderr, "WARN: Warning: Can't write system-time. Dump corrupted!\n");
  
  /* ---- Exit with an error ---- */
  exit(EXIT_FAILURE);

 }

}

void WriteSystemTimeHeader(PrgState *state)
{
 char charBuff;

#ifndef NDEBUG
 fprintf(stderr, 
         "DEBUG: Current system-time since epoch 0x%lX\n", 
         (long) state -> currentTime);
#endif

 /* ---- Convert the system time into a stream of hex-digits ---- */
 WriteTime(state -> outfd, state -> currentTime);

 /* ---- Write end-of-time marker ---- */
 charBuff = TIMEENDMARKER;
 if (write(state -> outfd, &(charBuff), sizeof(charBuff)) != sizeof(charBuff)) {

  /* ---- Tell that outputfile isn't complete ---- */
  fprintf(stderr, "WARN: Can't write end-of-system-time marker. Dump corrupted!\n");
  
  /* ---- Exit with an error ---- */
  exit(EXIT_FAILURE);

 }

}

void WriteProfile(PrgState* state)
{
 long i;
 long nwrite;
 unsigned long int checkSum = 0;

 /* ---- Check for verbose mode ---- */
 if (verbose) {

  /* ---- Check for real outputfile or standard-out as an outputfile ---- */
  if ((strcmp(state -> outfilename, "-") != 0) && 
      (state -> outfilename[0] != '\0')) {
   
   /* ---- Print the name of the outputfile ---- */   
   fprintf(stderr, "INFO: Write profile data to file %s\n", state -> outfilename);

   } else {

    /* ---- Tell that we write the standard-out ---- */
    fprintf(stderr, "INFO: Write profile data to standard output channel (stdout)\n");

   }

 }

 /* ---- Write the system-time in a portable way ---- */
 (void) WriteSystemTimeHeader(state);

 /* ---- Write handshake data to output file ---- */
 if ((nwrite = write(state -> outfd, 
                     state -> handShakeData, 
                     HANDSHAKESIZE)) != HANDSHAKESIZE) {

  /* ---- Tell that outputfile isn't complete ---- */
  fprintf(stderr, "WARN:  Dump is corrupted (System time is missing)\n");

  /* ---- In verbose mode give some additional info about the problem ---- */
  if (verbose)
   fprintf(stderr, "WARN: Wrote %ld bytes instead of %d bytes \n", nwrite, HANDSHAKESIZE);

 }

 /* ---- Write profile data to output file ---- */
 if ((nwrite = write(state -> outfd, 
                     &(state -> profileData), 
                     PROFILESIZE)) != PROFILESIZE) {

  /* ---- Tell that outputfile isn't complete ---- */
  fprintf(stderr, "WARN: Dump is corrupted (Some profile-data is missing)\n");

  /* ---- In verbose mode give some additional info about the problem ---- */
  if (verbose)
   fprintf(stderr, "Wrote %ld bytes instead of %d bytes\n", nwrite, PROFILESIZE);

 }

 /* ---- Calculate the 16-bit checksum of profile-data ---- */
 checkSum = 0;
 for (i = 0; i < DIVEDATASIZE; i++) {
   
   /* ----- Sum the actual byte and calculate the 16-bit modulo ---- */
  checkSum += (unsigned char) state -> profileData[i + PREHEADERSIZE];
  checkSum %= 65536;

 }

 /* ---- Check for correct checksum ---- */
 if (checkSum != 
     256 * ((unsigned char) state -> profileData[PREHEADERSIZE + DIVEDATASIZE +1]) + 
           ((unsigned char) state -> profileData[PREHEADERSIZE + DIVEDATASIZE])) {

  /* ---- Give a warning ---- */
  fprintf(stderr, "WARN: Checksum of dump is not correct! ");
  fprintf(stderr, "Maybe the dump is corrupted.\n");

 } else {

  /* ---- In verbose mode tell that the checksum is ok ---- */
  if (verbose)
   fprintf(stderr, "INFO: Checksum correct\n");

 }

 /* ---- Print some additional info ---- */
 fprintf(stderr, "\n");
 fprintf(stderr, "Time-interval: %d sec\n", (unsigned int) state -> handShakeData[5]);
 fprintf(stderr, "Battery: ");
 if ((int) state -> handShakeData[4] == BATTOK) 
  fprintf(stderr, "Ok\n");
 else
  fprintf(stderr, "Low!\n");
 fprintf(stderr, 
         "Serial-Number: %u\n", 
         (unsigned int)(256 * (unsigned char) state -> handShakeData[7]) + 
	                       (unsigned char) state -> handShakeData[6]);  
 fprintf(stderr,
         "Current internal Sensus-time (secs since last full restart): %lu\n",
         16777216 * (unsigned long) state -> handShakeData[11]  + 
            65526 * (unsigned long) state -> handShakeData[10]  + 
              256 * (unsigned long) state -> handShakeData[9] + 
                    (unsigned long) state -> handShakeData[8]); 
 fprintf(stderr, "\n");

 /* ---- Close outputfile (if != stdout) ---- */
 if (state -> outfd != STDOUT_FILENO) {

  /* ---- Close the output-file ---- */
  (void) close(state -> outfd);
  state -> outfd = -1;  

 }

}

int main (int argc, char* argv[])
{
 int key;
#ifdef HAVE_GETOPTLONG
 int optindex;
 struct option longargdesc[] = {
                                {"verbose", no_argument, NULL,      (int)'v'},
                                {"device",  required_argument, NULL,(int)'d'},
                                {"help",    no_argument, NULL,      (int)'h'},
                                {"output",  required_argument, NULL,(int)'o'},
                                {0,         0,                 NULL, 0}
                            };   
#endif

#ifndef NDEBUG
 /* ---- While debugging we are always verbose ---- */
 verbose = true;
#endif

 /* ---- Init internal program state variable ---- */
 InitState(&state);
 strncpy(state.outfilename, DEFOUTFILENAME, (size_t) (FILENAME_MAX - 1));

#ifdef HAVE_GETOPTLONG
 /* ---- Parse all options ---- */
 while ((key = getopt_long(argc, argv, "vhd:o:", longargdesc, &optindex)) != EOF) {
#else 
 /* ---- Parse all options ---- */
 while ((key = getopt(argc, argv, "vhd:o:")) != EOF) {
#endif
 
  /* ---- Check for specific option ---- */
  switch (key) {
 
   case 'v':
            /* ---- Set verbose mode ---- */
            verbose = true;
            break; 
   case 'h':
            /* ---- Print the help message ---- */
            PrintHelp(argv[0]);
            break;
   case 'd':
            /* ---- Check for verbose mode ---- */
            if (verbose)
             fprintf(stderr, "INFO: Set device to %s\n", optarg);

            /* ---- Copy new device name ---- */
            strncpy(state.devicename, optarg, (size_t) (FILENAME_MAX - 1));
            break;
 
   case 'o':
            /* ---- Check for verbose mode ---- */
            if (verbose)
             fprintf(stderr, "INFO: Set output to %s\n", optarg);

            /* ---- Copy name of output file ---- */
            strncpy(state.outfilename, optarg, (size_t) (FILENAME_MAX - 1));
            break;

   case '?':

#ifdef HAVE_GETLONG
            printf("Try `%s --help' for more information.\n", argv[0]);
#else
            printf("Try `%s -h' for more information.\n", argv[0]);
#endif
            exit(EXIT_SUCCESS);

   default:
            printf("Error while parsing options.\n");
            exit(EXIT_FAILURE);

  }

 }

 /* ---- We have not contacted the Sensus yet ---- */
 state.phase = PhaseNotContacted;

 /* ---- Install signal & exit - handler ---- */
 signal(SIGINT,  CleanUp);
 signal(SIGQUIT, CleanUp);
 signal(SIGTERM, CleanUp);
 (void) atexit(DoExit);

 /* ---- Create the lockfile and wait ---- */
 CreateLockfile(&state); MWait(LONGWAIT);

#ifndef NDEBUG
 fprintf(stderr, "\n\n");
 fprintf(stderr, "DEBUG: State after parsing the arguments and creating the lockfile:\n");
 PrintState(&state);
 fprintf(stderr, "\n\n");
#endif

 /* ---- Open the outputfile ---- */
 (void) HandleOutput(&state);

 /* ---- Open the device and set it to the needed mode (19200 8N1) ---- */
 (void) SetDevice(&state);

 /* ---- Start the handshake (See Sensus developer-manual) ---- */
 (void) MakeHandshake(&state);

 /* ---- The handshake was ok -> try to read the profile data ---- */
 (void) ReadProfile(&state);

 /* ---- Now all data is read -> write it to the output-file ---- */
 (void) WriteProfile(&state);

 /* --- Everything was fine (Call exit-handler) ---- */
 exit(EXIT_SUCCESS);

}
